import collections
import csv
import json
import os

import csv_stats


def item_to_tensor(col_name, stats, item):
    """Use stats[col_name] to tranform the item.

    Args:
      col_name: csv column name
      stats: dict of stats for all columns
      item: str value of the csv cell.

    Returns:
      converted cell as int / float. 
    """
    assert col_name in stats
    stat = stats[col_name]

    if stat['type'] == 'str':
        for idx, kv in enumerate(stat['str_stats']):
            if kv[0] == item:
                return idx
        return len(stat['str_stats'])
    elif stat['type'] == 'float':
        return (float(item) - stat['float_stats']['avg']) / stat['float_stats']['std']
    else:
        return (float(item) - stat['int_stats']['avg']) / stat['int_stats']['std']
            

def read_csv(fn, label_col, stats):
    """Convert data in csv to tensordict.

    Args:
      fn: name of the csv file
      label_col: column name of the label
      stats: Stats of the columns

    Returns:
      cols, tensor_dict, num_records tuple
      tensor_dict contains column name as key, list of float/int as value.
    """
    cols = []
    tensor_dict = collections.defaultdict(list)
    num_records = 0
    with open(fn) as inp:
        reader = csv.reader(inp)
        header = True
        for row in reader:
            if header:
                cols.extend(row)
                header = False
            else:
                label_found = False
                for idx, item in enumerate(row):
                    col_name = cols[idx]
                    if col_name == label_col:
                        tensor_dict[col_name].append(int(item))
                        label_found = True
                    else:
                        tensor_dict[col_name].append(
                            item_to_tensor(col_name, stats, item) * 1.0)
                num_records += 1
                if not label_found:
                    tensor_dict[label_col].append(0)

    return cols, tensor_dict, num_records


def csv_to_dataset(csv_fn, label_col, train_stats=None):
    if not train_stats:
        csvstats = csv_stats.CSVStats(csv_fn, label_col='outcome')
        train_stats = csvstats.getAllStats(verbose=True)

    cols, tensor_dict, num_records = read_csv(
        fn=csv_fn, label_col='outcome', stats=train_stats)

    train_dataset = []
    for idx in range(num_records):
        label = tensor_dict['outcome'][idx]
        one_record = []
        for col in cols:
            if col == 'outcome':
                continue
            one_record.append(tensor_dict[col][idx])
        train_dataset.append((one_record, label))

    return train_stats, train_dataset



import collections
import csv
import json
import numpy as np
import os
import scipy.stats


def genStats(items):
    """Given a list of str, generate stats.

    Args:
      items: list of str for a csv column.
    
    Returns:
      json dict of stat.
    """
    maybe_float = True
    maybe_int = True

    text_count = collections.defaultdict(int)
    float_vals = []
    int_vals = []
    text_vals = []
    for i in items:
        if maybe_float:
            try:
                float_val = float(i)
                float_vals.append(float_val)
            except Exception:
                # it is not a float
                maybe_float = False
                maybe_int = False
        if maybe_float and maybe_int:
            try:
                int_val = int(i)
                if int_val != float_val:
                    maybe_int = False
                else:
                    int_vals.append(int_val)
            except Exception:
                maybe_int = False
        text_vals.append(i)    
        text_count[i] += 1

    # sort text by counts
    text_counts = sorted(text_count.items(), key=lambda kv: kv[1], reverse=True)
    
    stats = {}
    if maybe_int:
        np_vals = np.array(int_vals)
        stats['type'] = 'int'
        stats['int_stats'] = {
            'avg': np.average(np_vals),
            'std': np.std(np_vals)
        }
        stats['raw'] = int_vals
    elif maybe_float:
        np_vals = np.array(float_vals)
        stats['type'] = 'float'
        stats['float_stats'] = {
            'avg': np.average(np_vals),
            'std': np.std(np_vals)
        }
        stats['raw'] = float_vals
    else:
        stats['type'] = 'str'
        stats['raw'] = text_vals
        stats['str_stats'] = text_counts

    return stats


def textToId(feature_stats):
    assert feature_stats['type'] == 'str'
    text_counts = feature_stats['str_stats']
    lookup = {}
    for idx, kv in enumerate(text_counts):
        lookup[kv[0]] = idx
    ids = []
    for text in feature_stats['raw']:
        ids.append(lookup[text])
    return ids


def getVals(feature_stats):
    if feature_stats['type'] == 'str':
        return textToId(feature_stats)
    return feature_stats['raw']
                       
    
def getCorr(feature_stats, label_stats):
    vals = np.array(getVals(feature_stats))
    labels = np.array(getVals(label_stats))
    return scipy.stats.pearsonr(vals, labels)


class CSVStats:
    """Read csv and gather stats for all columns."""
    def __init__(self, input_csv, label_col, save=False, stats_dir=None):
        self._input_csv = input_csv
        self._label_col = label_col
        self._save = save
        self._stats_dir = stats_dir
        if save:
            assert self._stats_dir
            

    def getAllStats(self, verbose=False):
        """Get stats for all columns.

        Args:
          output: set to print the stats.

        Returns:
          stats dict that contains stat for all columns.
        """
        cols = []
        data = collections.defaultdict(list)

        all_stats = {}

        with open(self._input_csv) as inp:
            reader = csv.reader(inp)
            header = True
            for row in reader:
                if header:
                    cols.extend(row)
                    header = False
                else:
                    for idx, item in enumerate(row):
                        data[cols[idx]].append(item)

        label_stats = genStats(data[self._label_col])

        for col in cols:
            stats = genStats(data[col])
            stats['corr'] = getCorr(stats, label_stats)
            if self._save:
                with open(os.path.join(self._stats_dir, col + '.json'), 'w') as out_f:
                    json.dump(stats, out_f)
            if verbose:
                stats_output = stats.copy()
                del stats_output['raw']
                print(col, stats_output)
            all_stats[col] = stats

        return all_stats
    



import json
import sklearn.metrics
import torch
import torch.nn as nn
import torch.nn.functional as F

class DenseLogisticRegression(nn.Module):
    def __init__(self, input_dim):
        super().__init__()
        self.fc1 = nn.Linear(input_dim, 50)
        self.relu1 = nn.ReLU()
        self.fc2 = nn.Linear(50, 20)
        self.relu2 = nn.ReLU()
        self.fc3 = nn.Linear(20, 1)        
        self.init_weights()

    def init_weights(self):
        initrange = 0.5
        self.fc1.weight.data.uniform_(-initrange, initrange)
        self.fc1.bias.data.zero_()
        self.fc2.weight.data.uniform_(-initrange, initrange)
        self.fc2.bias.data.zero_()
        self.fc3.weight.data.uniform_(-initrange, initrange)
        self.fc3.bias.data.zero_()

    def forward(self, inputs):
        fc1 = self.fc1(inputs)
        fc2 = self.fc2(self.relu1(fc1))
        fc3 = self.fc3(self.relu2(fc2))
        return torch.sigmoid(fc3)

    @classmethod
    def criterion(cls):
        return torch.nn.BCELoss(size_average=True)

    



import json
import sklearn.metrics
import time
import torch
import torch.nn as nn
import torch.nn.functional as F

from torch.utils.data import DataLoader


def generate_batch(batch):
    feature = torch.tensor([entry[0] for entry in batch], dtype=torch.float32)
    label = torch.tensor([entry[1] for entry in batch], dtype=torch.float32)
    label = label.unsqueeze(1)
    return feature, label


class Trainer():
    def __init__(self, model, train_data, dev_data):
        self.train_data = train_data
        self.dev_data = dev_data
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model = model.to(self.device)
        self.criterion = model.criterion()
        self.optimizer = torch.optim.SGD(self.model.parameters(), lr=1.0)
        self.scheduler = torch.optim.lr_scheduler.StepLR(self.optimizer, 1, gamma=0.9)

    def train(self, batch_size=128):
        # Train the model
        train_loss = 0
        train_acc = 0
        data = DataLoader(self.train_data, batch_size=batch_size, shuffle=True, collate_fn=generate_batch)
        for i, (inputs, cls) in enumerate(data):
            self.optimizer.zero_grad()
            inputs, cls = inputs.to(self.device), cls.to(self.device)
            output = self.model(inputs)
            self.loss = self.criterion(output, cls)
            self.loss.backward()
            self.optimizer.step()

            train_loss += self.loss.item()
            train_acc += ((output > 0.5) == cls).sum().item()

        # Adjust the learning rate
        self.scheduler.step()

        train_len = len(self.train_data)
        return train_loss / train_len, train_acc / train_len

    def dev(self, batch_size=128):
        loss = 0
        acc = 0
        data = DataLoader(self.dev_data, batch_size=batch_size, collate_fn=generate_batch)
        y = []
        pred = []
        for inputs, cls in data:
            y.extend(cls.tolist())
            inputs, cls = inputs.to(self.device), cls.to(self.device)

            with torch.no_grad():
                output = self.model(inputs)
                loss = self.criterion(output, cls)
                loss += loss.item()
                acc += ((output > 0.5) == cls).sum().item()
                pred.extend(output.cpu().tolist())

        auc_roc = sklearn.metrics.roc_auc_score(y, pred)
        dev_len = len(self.dev_data)
        return loss / dev_len, acc / dev_len, auc_roc

    def predict(self, infer_data, batch_size=128):
        data = DataLoader(infer_data, batch_size=batch_size, collate_fn=generate_batch)
        pred = []
        for inputs, _ in data:
            inputs = inputs.to(self.device)
            with torch.no_grad():
                output = self.model(inputs)
                pred.extend(output.cpu().tolist())

        return pred

    def train_loop(self, batch_size=128, n_epochs=50):
        min_valid_loss = float('inf')
        for epoch in range(n_epochs):
            start_time = time.time()
            train_loss, train_acc = self.train(batch_size)
            valid_loss, valid_acc, valid_aucroc = self.dev(batch_size)
            secs = int(time.time() - start_time)
            mins = secs / 60
            secs = secs % 60

            print('Epoch: %d, | time in {%d} minutes, {%d} seconds.' % (epoch + 1, mins, secs))
            print('\tLoss: %.4f(train)\t|\tAcc: %.1f%%(train)' % (train_loss, train_acc * 100))
            print('\tLoss: %.4f(test)\t|\tAcc: %.1f%%(test)|\tAuc: %.1f%%(test)' % (valid_loss, valid_acc * 100, valid_aucroc * 100))

    



import pandas as pd
import numpy as np

from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.base import BaseEstimator


class BasicTransformer(BaseEstimator):
    
    def __init__(self, cat_threshold=None, num_strategy='median', return_df=False):
        self.cat_threshold = cat_threshold
        if num_strategy not in ['mean', 'median']:
            raise ValueError('num_strategy must be either "mean" or "median"')
        self.num_strategy = num_strategy
        self.return_df = return_df
            
    def fit(self, X, y=None):
        self._columns = X.columns.values
        
        self._dtypes = X.dtypes.values
        self._kinds = np.array([dt.kind for dt in X.dtypes])
        self._column_dtypes = {}
        is_cat = self._kinds == 'O'
        self._column_dtypes['cat'] = self._columns[is_cat]
        self._column_dtypes['num'] = self._columns[~is_cat]
        self._feature_names = self._column_dtypes['num']
        
        self._cat_cols = {}
        for col in self._column_dtypes['cat']:
            vc = X[col].value_counts()
            if self.cat_threshold is not None:
                vc = vc[vc > self.cat_threshold]
            vals = vc.index.values
            self._cat_cols[col] = vals
            self._feature_names = np.append(self._feature_names, col + '_' + vals)
            
        self._total_cat_cols = sum([len(v) for col, v in self._cat_cols.items()])
        num_cols = self._column_dtypes['num']
        self._num_fill = X[num_cols].agg(self.num_strategy)
        return self
    
    def transform(self, X):
        # check that we have a DataFrame with same column names as 
        # the one we fit
        if set(self._columns) != set(X.columns):
            raise ValueError('Passed DataFrame has different columns than fit DataFrame')
        elif len(self._columns) != len(X.columns):
            raise ValueError('Passed DataFrame has different number of columns than fit DataFrame')
            
        # fill missing values
        num_cols = self._column_dtypes['num']
        X_num = X[num_cols].fillna(self._num_fill)
        
        # Standardize numerics
        std = X_num.std()
        X_num = (X_num - X_num.mean()) / std
        zero_std = np.where(std == 0)[0]
        
        # If there is 0 standard deviation, then all values are the 
        # same. Set them to 0.
        if len(zero_std) > 0:
            X_num.iloc[:, zero_std] = 0
        X_num = X_num.values
        
        # create separate array for new encoded categoricals
        X_cat = np.empty((len(X), self._total_cat_cols), 
                         dtype='int')
        i = 0
        for col in self._column_dtypes['cat']:
            vals = self._cat_cols[col]
            for val in vals:
                X_cat[:, i] = X[col] == val
                i += 1
                
        # concatenate transformed numeric and categorical arrays
        data = np.column_stack((X_num, X_cat))
        
        # return either a DataFrame or an array
        if self.return_df:
            return pd.DataFrame(data=data, 
                                columns=self._feature_names)
        else:
            return data
    
    def fit_transform(self, X, y=None):
        return self.fit(X).transform(X)
    
    def get_feature_names():
        return self._feature_names



